from .basic import JSComponent, CSSComponent, openfile
from .components import CoreComponent, Virtual

import pystache
import flask
import re
import os

from ..util import memoize

JSComponent.alias_requires("backbone", "vendor/backbone")

class SuperfluousLoader(CoreComponent, JSComponent):
    WRAP_COMPONENT = False

# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')
def to_snake_case(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()

@Virtual
class SuperfluousComponent(JSComponent, CSSComponent):
    @classmethod
    def get_file_for_ext(cls, ext):
        cname = to_snake_case(cls.__name__)
        cls_dir = cls.get_dir()
        fname = os.path.join(cls_dir, cname, "%s.%s" % (cname, ext))
        return fname

    @classmethod
    def get_class_dependencies(cls):
        return [ SuperfluousLoader ]

    @classmethod
    def get_template(cls):
        with openfile(cls.get_file_for_ext("html.erb")) as f:
            return f.read()

    @classmethod
    def get_events(cls):
        cname = to_snake_case(cls.__name__)
        cls_dir = cls.get_dir()
        with openfile(os.path.join(cls_dir, cname, "events.js")) as f:
            return f.read()

    @classmethod
    @memoize
    def get_package_object(cls):
        # in a superfluous package, it is:
        # the_component_name.html.erb for template
        # the_component_name.js for js
        # events.js for events
        # the_component_name.css for css
        ret = {}

        j = cls.get_js()
        js_supplements = cls.get_js_supplements()
        if js_supplements:
            js_supplements = "\n".join(js_supplements)

        ret["template"] = cls.get_template()
        ret["css"] = cls.get_css()
        ret["js"] = "%s\n%s" % (j, js_supplements or "")
        ret["requires"] = cls.get_requires()
        ret["defines"] = cls.get_defines()
        ret["events"] = cls.get_events()
        ret["__versions__"] = cls.get_require_versions()

        # clean up items
        ret = {k: v for k, v in ret.items() if v}

        return ret


    def __json__(self):
        self.__marshal__()
        return { "_S" : self.__html_id__() }

    def __activate__(self):
        self.client.update(self.context)
        super(SuperfluousComponent, self).__activate__()
        # TODO: SuperfluousLoader should be referenced via intermediate Class

        # we override the activation string with our backbone activation string
        t = """
            $P._load("SuperfluousLoader", function(m) {
                m.exports.activate_superfluous_component("{{__html_id__}}", "{{ __template_name__ }}", {{ &__context__ }}, {{ __display_immediately__ }}, "{{ __ref__ }}" )
            });
        """.strip()

        self.__activate_str__ = pystache.render(t, self)



from .components import *
from . import proxy

from .. import util

import flask
import addict
import json

class AssetLoader(Component):
    EXT="???"
    @classmethod
    def find_file(cls, fname, basedir, cls_dir):
        fname = fname.strip("'\"")
        if fname.endswith(cls.EXT):
            fname,ext = os.path.splitext(fname)

        if fname[0] == ".":
            jsp = "%s.%s" % (os.path.join(cls_dir, basedir, fname), cls.EXT)
        else:
            jsp = "%s.%s" % (os.path.join(cls_dir, fname), cls.EXT)

        return jsp

    @classmethod
    def render_file_to_js(cls, filename, error_if_missing=False):
        try:
            with util.open(filename, "r") as f:
                js = cls.transform_and_wrap_in_js(f.read())
                return js
        except IOError as e:
            if error_if_missing:
                raise e

        return ""

    @classmethod
    def match(cls, fname):
        return fname.endswith(".%s" % cls.EXT)

    @classmethod
    def transform(cls, js):
        return js

    @classmethod
    def transform_and_wrap_in_js(cls, js):
        dirhash = cls.get_dirhash()
        basehash = cls.get_basehash()
        return "require.__basehash = '%s'\nrequire.__dirhash = '%s';\n %s" % (basehash, dirhash, cls.transform(js))

class JSAsset(AssetLoader):
    EXT="js"


class CssAsset(AssetLoader):
    EXT="css"

    @classmethod
    def scopename(cls, css):
        import hashlib
        m = hashlib.md5()
        m.update(css.encode("utf-8"))

        hashed = m.hexdigest()[:8]

        return "scoped_" + hashed


    @classmethod
    def inject_css(cls,css, scope=""):
        if not scope:
            scope = cls.scopename(css)

        return """
            var _inj = %s;\n $P._inject_css(_inj.scope, _inj.css);
            module.exports.className = _inj.scope;

        """ % json.dumps({ "css": css, "scope": scope })

    @classmethod
    def get_definitions(cls):
        static_folder = flask.current_app.static_folder
        try:
            with util.open(os.path.join(static_folder, "definitions.sass")) as f:
                return f.read()
        except:
            return ""

    @classmethod
    def transform(cls, css, scope=""):
        d = cls.get_definitions()

        css = "%s\n%s" % (d, css)
        if scope:
            return sass.compile(string=".%s { %s }" % (scope, css))

        return sass.compile(string=css)

    @classmethod
    def transform_and_wrap_in_js(cls, css):
        scope = cls.scopename(css)
        css = cls.transform(css, scope=scope)
        return cls.inject_css(css, scope)


class SassAsset(CssAsset):
    EXT="sass"



import flask
import jinja2

# JinjaComponent: foo/foo.html
# BackboneComponent: foo/foo.js
# MustacheComponent: foo/foo.mustache
# SassComponent: foo/foo.css (runs it through sass pre-processor)
# CSSComponent:  foo/foo.css (doesn't run it through pre-processor)

# ClientBridge: requires foo/foo.js, exposes the client API on the server
# ServerBridge: requires foo/foo.js, exposes the server API on the client

# FlaskPage: takes template to render as named parameter
# BigJSPackage: renders the component as one large package instead of a split

from .components import Component, CoreComponent, set_base_dir, validate_components, list_components

# templating
from .basic import MustacheComponent, JinjaComponent
# style
from .basic import SassComponent, CSSComponent, BigCSSPackage
# javascripts
from .basic import JSComponent, BigJSPackage


from .bridge import ClientBridge, ServerBridge
from .bigpipe import Pipeline, Pagelet, NoJSPagelet
from .react import ReactComponent
from .backbone import BackboneComponent
from .superfluous import SuperfluousComponent
from .page import Page, FlaskPage

from .components import Virtual, get_basedir, get_baseclass, add_dirhash_alias



from .components import *
from .basic import JSComponent

import flask
import time
import struct
import zlib

# https://stackoverflow.com/questions/44185486/generate-and-stream-compressed-file-with-flask
def streaming_compress(generator):
    # Yield a gzip file header first.
    try:
        inttype = long
    except:
        inttype = int

    yield (
        b'\037\213\010\000' + # Gzip file, deflate, no filename
        struct.pack('<L', inttype(time.time())) +  # compression start time
        b'\002\377'  # maximum compression, no OS specified
    )

    # bookkeeping: the compression state, running CRC and total length
    compressor = zlib.compressobj(
        9, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
    crc = zlib.crc32(b"")
    length = 0

    for data in generator:
        data = data.encode("utf-8")
        chunk = compressor.compress(data)
        if chunk:
            yield chunk
        else:
            # note: do not use Z_FULL_FLUSH, it resets the compressor
            yield compressor.flush(zlib.Z_SYNC_FLUSH)

        crc = zlib.crc32(data, crc) & 0xffffffff
        length += len(data)

    yield compressor.flush()
    yield struct.pack("<2L", crc, length & 0xffffffff)


# This is the main loop for the preparer. It does a sleep spin
def sleep_spin(preparer, pagelets):
    start = time.time()
    pagelets = set(pagelets)
    while True:
        time.sleep(0.01)
        preparer.finish_tasks()

        remove = []
        for p in pagelets:
            if p.__finished__ and not p.__injected__:
                p.__injected__ = True
                remove.append(p)
                yield inject_pagelet(p)

        for p in remove:
            pagelets.remove(p)

        if preparer.done or not preparer.preparing:
            break

def inject_pagelet(pagelet):
    html = pagelet.render()
    t = "\n<div id='pagelet_%s'><!-- %s --></div>\n" % (pagelet.__html_id__(), flask.escape(html))
    i = "\n<script>$P._inject_pagelet('%s')</script>\n" % pagelet.__html_id__()
    return "\n".join([t, i])


class Pipeline(Component):
    def pipeline(self):

        flask.request.pudgy.pipelined = True



        stream =  flask.stream_with_context(self.run())
        accept_encoding = flask.request.headers.get('Accept-Encoding', '')
        if 'gzip' not in accept_encoding.lower():
            return flask.Response(stream)


        response = flask.Response(streaming_compress(stream))
        response.headers['Content-Encoding'] = 'gzip'

        return response

    def run(self):
        import preparable
        from ..blueprint import marshal_components

        kwargs = self.context
        r = self.render()

        yield r
        yield marshal_components(prelude=True)

        preparer = preparable.Preparer()
        pagelets = []
        for p in flask.request.pudgy.pagelets:
            if not p.__async__:
                continue

            p.__done__ = False
            preparer.add(p.__prepare_pagelet__, [])
            pagelets.append(p)


        preparer.startup()
        for p in sleep_spin(preparer, pagelets):
            yield p

        for p in flask.request.pudgy.pagelets:
            if not p.__async__:
                continue

            flask.request.pudgy.components.add(p)

            if not p.__injected__:
                yield inject_pagelet(p)

        yield marshal_components(prelude=False)



class Pagelet(JSComponent):
    def __init__(self, *args, **kwargs):
        self.__finished__ = False
        self.__async__ = False
        super(Pagelet, self).__init__(*args, **kwargs)

        if flask.request:
            flask.request.pudgy.pagelets.add(self)

    def async(self):
        self.__async__ = True

    def render(self):
        if self.__async__ and not self.__finished__:
            t = """<div class='async_pagelet pagelet' id='pl_{{ id }}'> </div>"""
            r = pystache.render(t, { "id" : self.__html_id__() })
            return r
        else:
            return super(Pagelet, self).render()

    def __prepare_pagelet__(self):
        self.__injected__ = False
        import types
        r = self.__prepare__()

        if isinstance(r, types.GeneratorType):
            for p in r:
                yield p

            self.__finished__ = True
        else:
            self.__finished__ = True
            yield r


    def __activate_tag__(self):
        if self.__async__ and not self.__finished__:
            return ""

        return super(Pagelet, self).__activate_tag__()

    def __work__(self):
        pass

class NoJSPagelet(Pagelet):
    # has no JS file associated with it
    @classmethod
    def get_js(cls):
        return ""

mark_virtual(Pagelet, Pipeline, NoJSPagelet)



import hashlib
import time
import os

import flask
import io

import sys
if sys.version_info[0] >= 3:
    unicode = str


def open(fname, mode="r", encoding="utf-8"):
    return io.open(fname, mode, encoding=encoding)

def inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses

# from https://stackoverflow.com/questions/16463582/memoize-to-disk-python-persistent-memoization
# https://stackoverflow.com/posts/47385932/revisions
from diskcache import Cache
cache_dir = "./cache/"
def shelve_it(table_name):
    cache_file = os.path.join(cache_dir, table_name)
    d = Cache(cache_file)

    def decorator(func):
        def new_func(*args, **kwargs):
            cache_key = "\r".join([str(arg) for arg in args])

            if cache_key in d:
                return d[cache_key]


            ret = func(*args, **kwargs)
            d[cache_key] = ret

            return ret

        return new_func

    return decorator

def memoize(func):
    cache = {}
    def new_func(*args, **kwargs):
        cache_key = "\r".join([str(arg) for arg in args])
        if kwargs:
            cache_key += "\r".join(["%s:%s" % (k,v) for k,v in kwargs.iteritems()])

        if cache_key in cache:
            return cache[cache_key]

        ret = func(*args, **kwargs)
        cache[cache_key] = ret
        return ret

    new_func.cache = cache
    return new_func

def gethash(v):
    m = hashlib.md5()
    h = unicode(v).strip()
    u = h.encode("utf-8")

    m.update(u)

    return m.hexdigest()

def getrandhash(v):
    m = hashlib.md5()
    t = str("%s" % time.time()).encode("utf-8")
    h = str(hash(v)).encode("utf-8")

    m.update(t)
    m.update(h)

    return m.hexdigest()

def handle_request_to(endpoint, **values):
    url = flask.url_for(endpoint, **values)
    req = flask.Request.from_values(path=url)
    ctx = flask.current_app.test_request_context(path=url)

    with ctx:
        res = flask.current_app.dispatch_request()
    return res

CACHEABLE_ASSET_ENDPOINTS = [
    'components.get_prelude',
    'components.get_big_css',
    'components.get_component_specs' ]
def dated_url_for(endpoint, **values):
    app = flask.current_app
    # we know prelude always returns a string, so
    # we use python string hashing
    if endpoint in CACHEABLE_ASSET_ENDPOINTS:
        res = handle_request_to(endpoint, **values)
        hsh = gethash(res)

        values['q'] = "%s" % (hsh[:10])

    # if the endpoint is a filename, we use timestamp hashing
    # (but we should really use content hashing, i guess)
    if endpoint == 'static':
        filename = values.get('filename', None)
        if filename:
            file_path = os.path.join(app.root_path,
                                     endpoint, filename)
            values['q'] = int(os.stat(file_path).st_mtime)

    return flask.url_for(endpoint, **values)




from .blueprint import register_blueprint, compress_request
from .components.exported import *

from . import util

from .prelude import add_to_prelude, add_prelude_line, use_jquery, preload_components
